/**************************************************************************
 *
 * Copyright 2016 Samsung Electronics co., Ltd. All Rights Reserved.
 *
 * Contact: Konstantin Drabeniuk <k.drabeniuk@samsung.com>
 * Contact: Andrii Sokolenko <a.sokolenko@samsung.com>
 * Contact: Roman Marchenko <r.marchenko@samsung.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sub license, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
 * IN NO EVENT SHALL PRECISION INSIGHT AND/OR ITS SUPPLIERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
**************************************************************************/

#include "tc_tbm.h"

class TBMSurfaceInternal : public TBMBufmgr
{
protected:
	void SetUp();
	void TearDown();
public:
	uint32_t num_formats;
	tbm_format *formats;
	tbm_format argb8888_format;

	tbm_surface_h surface;
};

void TBMSurfaceInternal::SetUp()
{
	int ret = 0;

	TBMBufmgr::SetUp();

	ret = tbm_surface_internal_query_supported_formats(&formats, &num_formats);
	ASSERT_TRUE(ret == 1);

	for (uint32_t i = 0; i < num_formats; i++)
		if (formats[i] == TBM_FORMAT_ARGB8888)
			argb8888_format = TBM_FORMAT_ARGB8888;

	/* fail if there is no TBM_FORMAT_ARGB8888 format */
	ASSERT_TRUE(argb8888_format == TBM_FORMAT_ARGB8888);

	surface = tbm_surface_internal_create_with_flags(720, 1024, argb8888_format, TBM_BO_DEFAULT);
	ASSERT_TRUE(surface != NULL);
}

void TBMSurfaceInternal::TearDown()
{
	if (surface)
		tbm_surface_internal_destroy(surface);

	if (formats)
		free(formats);

	TBMBufmgr::TearDown();
}

static int memory_for_invalid_param;
static tbm_surface_h invalid_surface = (tbm_surface_h)&memory_for_invalid_param;

/* tbm_surface_h tbm_surface_internal_create_with_flags(int width, int height, int format, int flags) */
TEST_F(TBMSurfaceInternal, SurfaceInternalCeateDestroyTest)
{
	tbm_surface_h surface1 = NULL;

	for (uint32_t i = 0; i < 1; i++) {
		surface1 = tbm_surface_internal_create_with_flags(720, 1024, formats[i], TBM_BO_DEFAULT);
		EXPECT_NE(surface1, nullptr);
		EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
		tbm_surface_internal_destroy(surface1);
		EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	}
}

TEST_F(TBMSurfaceInternal, SurfaceInternalCeatewithParamsTest)
{
	tbm_surface_h surface1 = NULL;

	surface1 = tbm_surface_internal_create_with_flags(-1, 1024, TBM_FORMAT_ARGB8888, TBM_BO_DEFAULT);
	EXPECT_EQ(nullptr, surface1); // Expected Value: NULL
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
	surface1 = tbm_surface_internal_create_with_flags(720, -1, TBM_FORMAT_ARGB8888, TBM_BO_DEFAULT);
	EXPECT_EQ(nullptr, surface1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
	surface1 = tbm_surface_internal_create_with_flags(720, 1024, 0, TBM_BO_DEFAULT);
	EXPECT_EQ(nullptr, surface1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
	surface1 = tbm_surface_internal_create_with_flags(720, 1024, 0, (TBM_BO_WC << 1));
	EXPECT_EQ(nullptr, surface1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

/* void tbm_surface_internal_destroy(tbm_surface_h surface) */
TEST_F(TBMSurfaceInternal, SurfaceInternalSurfaceInteranlDestroyWithParamsTest)
{
	tbm_surface_internal_destroy(NULL);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, SurfaceInteranlDestroyTwiceTest)
{
	tbm_surface_h surface1 = NULL;

	surface1 = tbm_surface_internal_create_with_flags(720, 1024, TBM_FORMAT_ARGB8888, TBM_BO_DEFAULT);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	tbm_surface_internal_destroy(surface1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	tbm_surface_internal_destroy(surface1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

/* tbm_surface_internal_is_valid(tbm_surface_h surface) */
TEST_F(TBMSurfaceInternal, SurfaceInteranlIsValid)
{
	EXPECT_TRUE(tbm_surface_internal_is_valid(surface));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	EXPECT_FALSE(tbm_surface_internal_is_valid(NULL));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	EXPECT_FALSE(tbm_surface_internal_is_valid((tbm_surface_h)&invalid_surface));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

/* int tbm_surface_internal_get_num_bos(tbm_surface_h surface) */
TEST_F(TBMSurfaceInternal, GetNumBosFailNULL)
{
	int num = tbm_surface_internal_get_num_bos(NULL);
	EXPECT_EQ(0, num);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, GetNumBosSuccess)
{
	int num = tbm_surface_internal_get_num_bos(surface);
	EXPECT_GT(num, 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
}

/* tbm_bo tbm_surface_internal_get_bo(tbm_surface_h surface, int bo_idx) */
TEST_F(TBMSurfaceInternal, GetBoFailNULL)
{
	tbm_bo bo = tbm_surface_internal_get_bo(NULL, 0);
	EXPECT_EQ(nullptr, bo);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, GetBoFailInvalideInput)
{
	tbm_bo bo = NULL;

	/* invalid surface */
	bo = tbm_surface_internal_get_bo(invalid_surface, 0);
	EXPECT_EQ(nullptr, bo);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	/* invalid index of bo */
	bo = tbm_surface_internal_get_bo(surface, -1);
	EXPECT_EQ(nullptr, bo);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, GetBoSuccess)
{
	tbm_bo bo = NULL;
	int num = 0;

	num = tbm_surface_internal_get_num_bos(surface);
	EXPECT_GT(num, 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	for (int i = 0; i < num; i++) {
		bo = tbm_surface_internal_get_bo(surface, i);
		EXPECT_NE(nullptr, bo);
		EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	}
}

/* tbm_surface_h tbm_surface_internal_create_with_bos(tbm_surface_info_s *info, tbm_bo *bos, int num) */
TEST_F(TBMSurfaceInternal, CreateWithBosNullInput)
{
	tbm_surface_h surf = NULL;
	tbm_surface_info_s info = { 0 };
	tbm_surface_error_e result = TBM_SURFACE_ERROR_NONE;
	tbm_bo bos[1];

	result = (tbm_surface_error_e)tbm_surface_get_info(surface, &info);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_NONE));

	surf = tbm_surface_internal_create_with_bos(NULL, NULL, 1);
	EXPECT_EQ(nullptr, surf);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	surf = tbm_surface_internal_create_with_bos(&info, NULL, 1);
	EXPECT_EQ(nullptr, surf);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	surf = tbm_surface_internal_create_with_bos(NULL, bos, 1);
	EXPECT_EQ(nullptr, surf);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, CreateWithBosFailInvalidInput)
{
	tbm_surface_h surf = NULL;
	tbm_surface_info_s info;
	tbm_surface_error_e result = TBM_SURFACE_ERROR_NONE;
	tbm_bo bos[4] = { NULL };
	int num;

	result = (tbm_surface_error_e)tbm_surface_get_info(surface, &info);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_NONE));

	num = tbm_surface_internal_get_num_bos(surface);
	EXPECT_GT(num, 0);

	for (int i = 0; i < num; i++) {
		bos[i] = tbm_surface_internal_get_bo(surface, i);
		EXPECT_NE(nullptr, bos[i]);
		EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	}

	/* wrong number of bo */
	surf = tbm_surface_internal_create_with_bos(&info, bos, num + 1);
	EXPECT_EQ(nullptr, surf);
	EXPECT_STRNE(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	surf = tbm_surface_internal_create_with_bos(&info, bos, num - 1);
	EXPECT_EQ(nullptr, surf);
	EXPECT_STRNE(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	/* invalid bo */
	bos[num-1] = NULL;
	surf = tbm_surface_internal_create_with_bos(&info, bos, num);
	EXPECT_EQ(nullptr, surf);
	EXPECT_STRNE(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
}

TEST_F(TBMSurfaceInternal, CreateWithBosSuccess)
{
	tbm_surface_h surf = NULL;
	tbm_surface_info_s info;
	tbm_surface_error_e result = TBM_SURFACE_ERROR_NONE;
	tbm_bo bos[4] = { NULL };
	int num = 0;

	result = (tbm_surface_error_e)tbm_surface_get_info(surface, &info);
	EXPECT_EQ(TBM_SURFACE_ERROR_NONE, result);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_NONE));

	num = tbm_surface_internal_get_num_bos(surface);
	EXPECT_GT(num , 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	for (int i = 0; i < num; i++) {
		bos[i] = tbm_surface_internal_get_bo(surface, i);
		EXPECT_NE(nullptr, bos[i]);
		EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	}

	surf = tbm_surface_internal_create_with_bos(&info, bos, num);
	EXPECT_NE(nullptr, surf);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	tbm_surface_internal_destroy(surf);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	/* set size plane to zero*/
	info.planes[0].size = 0;
	surf = tbm_surface_internal_create_with_bos(&info, bos, num);
	EXPECT_NE(nullptr, surf);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	tbm_surface_internal_destroy(surf);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	/* set bpp to zero*/
	info.bpp = 0;
	surf = tbm_surface_internal_create_with_bos(&info, bos, num);
	EXPECT_NE(nullptr, surf);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	tbm_surface_internal_destroy(surf);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	/* set size to zero*/
	info.size = 0;
	surf = tbm_surface_internal_create_with_bos(&info, bos, num);
	EXPECT_NE(nullptr, surf);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	tbm_surface_internal_destroy(surf);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
}

/* void tbm_surface_internal_unref(tbm_surface_h surface) */
TEST_F(TBMSurfaceInternal, UnrefFailNull)
{
	tbm_surface_internal_unref(NULL);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, UnrefFailInvaildInput)
{
	int surf;
	tbm_surface_internal_unref((tbm_surface_h)&surf);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, UnrefSuccess)
{
	tbm_surface_internal_unref(surface);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	EXPECT_FALSE(tbm_surface_internal_is_valid(surface)) ;

	surface = NULL;
}

/* void tbm_surface_internal_ref(tbm_surface_h surface) */
TEST_F(TBMSurfaceInternal, RefFailNull)
{
	tbm_surface_internal_ref(NULL);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, RefFailInvaildInput)
{
	int surf;
	tbm_surface_internal_ref((tbm_surface_h)&surf);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, RefSuccess)
{
	tbm_surface_internal_ref(surface);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	tbm_surface_internal_unref(surface);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	EXPECT_TRUE(tbm_surface_internal_is_valid(surface));
}

/* tbm_surface_internal_get_num_planes() */
TEST_F(TBMSurfaceInternal, GetNumPlanes)
{
	int num = 0;
	num = tbm_surface_internal_get_num_planes(0);
	EXPECT_EQ(0, num);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	num = tbm_surface_internal_get_num_planes(TBM_FORMAT_ARGB8888);
	EXPECT_EQ(1, num);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	num = tbm_surface_internal_get_num_planes(TBM_FORMAT_NV12);
	EXPECT_EQ(2, num);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	num = tbm_surface_internal_get_num_planes(TBM_FORMAT_YUV410);
	EXPECT_EQ(3, num);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
}

/* tbm_surface_internal_get_size() */
TEST_F(TBMSurfaceInternal, GetSizeFailNull)
{
	int size = tbm_surface_internal_get_size(NULL);
	EXPECT_EQ(0, size);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, GetSizeFailInvaildInput)
{
	int surf;
	int size = tbm_surface_internal_get_size((tbm_surface_h)&surf);
	EXPECT_EQ(0, size);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, GetSizeSuccess)
{
	int size = tbm_surface_internal_get_size(surface);
	ASSERT_GT(size, 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
}

/* tbm_surface_internal_get_plane_data(tbm_surface_h surface, int plane_idx, uint32_t *size, uint32_t *offset, uint32_t *pitch) */
TEST_F(TBMSurfaceInternal, GetPlaneDataFailNull)
{
	EXPECT_FALSE(tbm_surface_internal_get_plane_data(NULL, 0, NULL, NULL, NULL));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, GetPlaneDataFailInvaildInput)
{
	int surf;

	EXPECT_FALSE(tbm_surface_internal_get_plane_data((tbm_surface_h)&surf, 0, NULL, NULL, NULL));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
	EXPECT_FALSE(tbm_surface_internal_get_plane_data(surface, -1, NULL, NULL, NULL));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
	EXPECT_FALSE(tbm_surface_internal_get_plane_data(surface, 3, NULL, NULL, NULL));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, GetPlaneDataSuccess)
{
	uint32_t size = 0, offset = 0, pitch = 0;

	EXPECT_TRUE(tbm_surface_internal_get_plane_data(surface, 0, &size, NULL, NULL));
	EXPECT_GT(size, 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	EXPECT_TRUE(tbm_surface_internal_get_plane_data(surface, 0, NULL, &offset, NULL));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	EXPECT_TRUE(tbm_surface_internal_get_plane_data(surface, 0, NULL, NULL, &pitch));
	EXPECT_GT(pitch, 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
}

/* tbm_surface_internal_get_bpp(tbm_format format) */
TEST_F(TBMSurfaceInternal, GetBpp)
{
	int bpp = 0;
	bpp = tbm_surface_internal_get_bpp(0);
	EXPECT_EQ(0, bpp);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	bpp = tbm_surface_internal_get_bpp(TBM_FORMAT_RGB332);
	EXPECT_EQ(8, bpp);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	bpp = tbm_surface_internal_get_bpp(TBM_FORMAT_RGB565);
	EXPECT_EQ(16, bpp);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	bpp = tbm_surface_internal_get_bpp(TBM_FORMAT_RGB888);
	EXPECT_EQ(24, bpp);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	bpp = tbm_surface_internal_get_bpp(TBM_FORMAT_ARGB8888);
	EXPECT_EQ(32, bpp);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	bpp = tbm_surface_internal_get_bpp(TBM_FORMAT_NV12);
	EXPECT_EQ(12, bpp);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	bpp = tbm_surface_internal_get_bpp(TBM_FORMAT_NV16);
	EXPECT_EQ(16, bpp);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	bpp = tbm_surface_internal_get_bpp(TBM_FORMAT_YUV410);
	EXPECT_EQ(9, bpp);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	bpp = tbm_surface_internal_get_bpp(TBM_FORMAT_YUV411);
	EXPECT_EQ(12, bpp);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	bpp = tbm_surface_internal_get_bpp(TBM_FORMAT_YUV422);
	EXPECT_EQ(16, bpp);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	bpp = tbm_surface_internal_get_bpp(TBM_FORMAT_YUV444);
	EXPECT_EQ(24, bpp);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
}
/* tbm_surface_internal_get_plane_bo_idx(tbm_surface_h surface, int plane_idx) */
TEST_F(TBMSurfaceInternal, GetPlaneBoIndexFailNull)
{
	int idx = tbm_surface_internal_get_plane_bo_idx(NULL, 0);
	EXPECT_EQ(0, idx);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, GetPlaneBoIndexFailInvaildInput)
{
	int surf;
	int idx;

	idx = tbm_surface_internal_get_plane_bo_idx((tbm_surface_h)&surf, 0);
	EXPECT_EQ(0, idx);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	idx = tbm_surface_internal_get_plane_bo_idx(surface, -1);
	EXPECT_EQ(0, idx);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, GetPlaneBoIndexSuccess)
{
	int idx;

	idx = tbm_surface_internal_get_plane_bo_idx(surface, 0);
	EXPECT_EQ(0, idx);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
}

static int tbm_data_free_is_called = 0;
void data_free_handle(void *user_data)
{
	ASSERT_TRUE(user_data == &tbm_data_free_is_called);
	int *is_called = (int *)user_data;
	*is_called = 1;
}

/* tbm_surface_internal_add_user_data(tbm_surface_h surface, unsigned long key, tbm_data_free data_free_func) */
TEST_F(TBMSurfaceInternal, AddUserDataFailNull)
{
	static const unsigned long key_ud = 0;
	EXPECT_EQ(tbm_surface_internal_add_user_data(NULL, key_ud, NULL), 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, AddUserDataFailInvaildInput)
{
	static const unsigned long key_ud = 0;
	int invalid_surf;
	EXPECT_EQ(tbm_surface_internal_add_user_data((tbm_surface_h)&invalid_surf, key_ud, NULL), 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, AddUserDataSuccess)
{
	unsigned long key_ud = 157;
	EXPECT_EQ(tbm_surface_internal_add_user_data(surface, key_ud, NULL), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
}

TEST_F(TBMSurfaceInternal, AddUserDataFailDoubleKey)
{
	unsigned long key_ud = 5487;
	EXPECT_EQ(tbm_surface_internal_add_user_data(surface, key_ud, NULL), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	EXPECT_EQ(tbm_surface_internal_add_user_data(surface, key_ud, NULL), 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

/* tbm_surface_internal_set_user_data(tbm_surface_h surface, unsigned long key, void *data) */
TEST_F(TBMSurfaceInternal, SetUserDataFailNull)
{
	static const unsigned long key_ud = 0;
	EXPECT_EQ(tbm_surface_internal_set_user_data(NULL, key_ud, &tbm_data_free_is_called), 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, SetUserDataFailInvaildInput)
{
	static const unsigned long key_ud = 0;
	int surf;

	EXPECT_EQ(tbm_surface_internal_set_user_data(surface,
			key_ud, &tbm_data_free_is_called), 0) << "invalid key";
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	EXPECT_EQ(tbm_surface_internal_set_user_data((tbm_surface_h)&surf,
			key_ud, &tbm_data_free_is_called), 0) << "invalid surface";
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, SetUserDataSuccess)
{
	unsigned long key_ud = 87947;
	EXPECT_EQ(tbm_surface_internal_add_user_data(surface, key_ud, data_free_handle), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	/*the first call*/
	EXPECT_EQ(tbm_surface_internal_set_user_data(surface, key_ud, &tbm_data_free_is_called), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	/*the second call: previous data has to be released*/
	tbm_data_free_is_called = 0;
	EXPECT_EQ(tbm_surface_internal_set_user_data(surface, key_ud, NULL), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	EXPECT_EQ(tbm_data_free_is_called, 1);

	/*the third call: previous data(NULL) hasn't to be released*/
	tbm_data_free_is_called = 0;
	EXPECT_EQ(tbm_surface_internal_set_user_data(surface, key_ud, NULL), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	EXPECT_EQ(tbm_data_free_is_called, 0);
}

/* tbm_surface_internal_get_user_data(tbm_surface_h surface, unsigned long key, void **data) */
TEST_F(TBMSurfaceInternal, GetUserDataFailNull)
{
	static const unsigned long key_ud = 0;
	int *data = NULL;
	EXPECT_EQ(tbm_surface_internal_get_user_data(NULL, key_ud, (void **)&data), 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	EXPECT_EQ(tbm_surface_internal_get_user_data(surface, key_ud, NULL), 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, GetUserDataFailInvaildInput)
{
	static const unsigned long key_ud = 0;
	int surf;
	int *data = NULL;

	EXPECT_EQ(tbm_surface_internal_get_user_data(surface,
			key_ud, (void **)&data), 0) << "invalid key" << std::endl;
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	EXPECT_EQ(tbm_surface_internal_get_user_data((tbm_surface_h)&surf,
			key_ud, (void **)&data), 0) << "invalid surface" << std::endl;
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, GetUserDataSuccess)
{
	unsigned long key_ud = 97456789;
	int *data = NULL;

	EXPECT_EQ(tbm_surface_internal_add_user_data(surface, key_ud, data_free_handle), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	EXPECT_EQ(tbm_surface_internal_set_user_data(surface, key_ud, &tbm_data_free_is_called), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	/* check set data */
	EXPECT_EQ(tbm_surface_internal_get_user_data(surface, key_ud, (void **)&data), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	EXPECT_EQ(data == &tbm_data_free_is_called, 1);

	/* check if data is NULL */
	EXPECT_EQ(tbm_surface_internal_set_user_data(surface, key_ud, NULL), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	EXPECT_EQ(tbm_surface_internal_get_user_data(surface, key_ud, (void **)&data), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	EXPECT_EQ(data == NULL, 1);
}

/* tbm_surface_internal_delete_user_data(tbm_surface_h surface, unsigned long key) */
TEST_F(TBMSurfaceInternal, DeleteUserDataFailNull)
{
	static const int key_ud = 0;
	EXPECT_EQ(tbm_surface_internal_delete_user_data(NULL, key_ud), 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, DeleteUserDataFailInvaildInput)
{
	static const int key_ud = 0;
	int surf;
	EXPECT_EQ(tbm_surface_internal_delete_user_data(surface, key_ud), 0) << "invalid key";
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	EXPECT_EQ(tbm_surface_internal_delete_user_data((tbm_surface_h)&surf, key_ud), 0) << "invalid surface";
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, DeleteUserDataSuccess)
{
	unsigned long key_ud = UINT_MAX;

	EXPECT_EQ(tbm_surface_internal_add_user_data(surface, key_ud, data_free_handle), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	EXPECT_EQ(tbm_surface_internal_set_user_data(surface, key_ud, &tbm_data_free_is_called), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	tbm_data_free_is_called = 0;
	EXPECT_EQ(tbm_surface_internal_delete_user_data(surface, key_ud), 1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
	EXPECT_EQ(tbm_data_free_is_called, 1);
	EXPECT_EQ(tbm_surface_internal_set_user_data(surface, key_ud, &tbm_data_free_is_called), 0);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurfaceInternal, ExportNegative_InvalidParameters)
{
	tbm_error_e error = TBM_ERROR_NONE;
	tbm_surface_buffer_data *buffer_data = NULL;

	// invalid surface test
	buffer_data = tbm_surface_internal_export(invalid_surface, &error);
	// skip this test if this api is not supported.
	TBM_UT_RETURN_TRUE_NOT_SUPPORTED(error);

	EXPECT_STREQ(tbm_error_str(error), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
	EXPECT_EQ(buffer_data, nullptr);
}

TEST_F(TBMSurfaceInternal, ExportPositive_Success)
{
	tbm_error_e error = TBM_ERROR_NONE;
	tbm_surface_buffer_data *buffer_data = NULL;
	unsigned int i;

	buffer_data = tbm_surface_internal_export(surface, &error);
	// skip this test if this api is not supported.
	TBM_UT_RETURN_TRUE_NOT_SUPPORTED(error);

	EXPECT_STREQ(tbm_error_str(error), tbm_error_str(TBM_ERROR_NONE));
	EXPECT_NE(buffer_data, nullptr);
	// verify the member variables of tbm_surface_buffer_data
	if (buffer_data) {
		EXPECT_GT(buffer_data->num_fds, 0);
		for (i = 0; i < buffer_data->num_fds; i++) {
			EXPECT_GT(buffer_data->fds[i], -1);
		}
	}
}

TEST_F(TBMSurfaceInternal, ImportNegative_InvalidParameters)
{
	tbm_error_e error = TBM_ERROR_NONE;
	tbm_surface_h surface1 = NULL;

	// invalid surface test
	surface1 = tbm_surface_internal_import(NULL, NULL, &error);
	TBM_UT_RETURN_TRUE_NOT_SUPPORTED(error);

	EXPECT_STREQ(tbm_error_str(error), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
	EXPECT_EQ(surface1, nullptr);
}

TEST_F(TBMSurfaceInternal, ImportPositive_Success)
{
	tbm_error_e error = TBM_ERROR_NONE;
	tbm_surface_buffer_data *buffer_data = NULL;
	tbm_surface_info_s surface_info;
	tbm_surface_h surface1 = NULL;
	unsigned int i;

	// export surface_data
	buffer_data = tbm_surface_internal_export(surface, &error);
	TBM_UT_RETURN_TRUE_NOT_SUPPORTED(error);

	EXPECT_STREQ(tbm_error_str(error), tbm_error_str(TBM_ERROR_NONE));
	EXPECT_NE(buffer_data, nullptr);

	// verify the member variables of tbm_surface_buffer_data
	if (buffer_data) {
		EXPECT_GT(buffer_data->num_fds, 0);
		for (i = 0; i < buffer_data->num_fds; i++) {
			EXPECT_GT(buffer_data->fds[i], -1);
		}
	}

	// get surface_info
	error = (tbm_error_e)tbm_surface_get_info(surface, &surface_info);
	EXPECT_EQ(TBM_SURFACE_ERROR_NONE, error);
	EXPECT_STREQ(tbm_error_str(error), tbm_error_str(TBM_ERROR_NONE));

	// import surface_data
	EXPECT_NE((surface1 = tbm_surface_internal_import(&surface_info, buffer_data, &error)), nullptr);
	EXPECT_STREQ(tbm_error_str(error), tbm_error_str(TBM_ERROR_NONE));

	tbm_surface_internal_destroy(surface1);
}
